/*******************************************************************************
* Copyright (C) NEC Electronics Corporation 2008
* NEC ELECTRONICS CONFIDENTIAL AND PROPRIETARY
* All rights reserved by NEC Electronics Corporation.
* Use of copyright notice does not evidence publication
*
*  File Name : emmc/emmc.c
*  Contents  : eMMC boot main
*  History   :
*   Date        Rev.      Who         Details     Keyword
*  ==========  ========  ==========  ==========  ==========
*  2009/8/19   Rev1.01   j-nakamu
*
*******************************************************************************/

#include <config.h>

#if	defined(CONFIG_MP200_EMMCBOOT)

#include "../include/mini_config.h"
#include "emmc.h"

/* static function (mini-boot & u-boot) */
static int emmc_send_cmd( int cmd );
static int emmc_multi_block_read( unsigned int address, unsigned char *data, unsigned int size);

#if defined(MP200_MINIBOOT)
/* #define MP200_EMMC_ACCMODE	1*/

/* global function */
void mini_boot( void );

/* extern function */
extern void uart_init(void);
extern void dbg_print(const char* s);

#else	/* u-boot */
#define MP200_EMMC_OK			0
#define MP200_EMMC_ERR_PARAM1	-1
#define MP200_EMMC_ERR_PARAM2	-2
#define MP200_EMMC_ERR_PARAM3	-3
#define MP200_EMMC_ERR_PROTECT	-4
#define MP200_EMMC_ERR_WRITE	-5
#define MP200_EMMC_ERR_READ		-6

#define MP200_ERASE_MAXBYTE		(0x80000000)	/* 2G Byte */
#define MP200_ERASE_MAXBLK		(MP200_ERASE_MAXBYTE / SDM_BLOCKLEN_VAL)	/* 4M Block */

/* boot information */
struct boot_info {
	unsigned long addr;		/* ddr address */
	unsigned long block;	/* block number */
	unsigned long size;		/* size [byte] */
	unsigned long erase;	/* erase flag */
};

/* static function definition */
static int emmc_check_protect( unsigned long, unsigned long, int );
static int emmc_multi_block_write( unsigned int address, unsigned char *data, unsigned int size);
static int emmc_multi_block_erase( unsigned int address, unsigned int size );

/* global variable */
int emmc_erase_blocks = CFG_EMMC_SECT_OF_BLOCKS;

/* static variable */
static int acc_mode = 0;

unsigned long 
emmc_flash_init( void )
{
	int				i, ret, block_num;
	unsigned short	tmp_u16;
	unsigned int	*wp_tbl = (unsigned int *)CFG_EMMC_PROTECT_TBL;
	MBRecord		*mbr_data = (MBRecord *)ROMBOOT_DATA_START;
	
	acc_mode = *(int *)SRAM_ACC_MODE;
	
	/* CMD7 : DESELECT_CARD */
	outl( 0, SDM_ARG0 );
	outl( 0, SDM_ARG1 );
	ret = emmc_send_cmd( 7 );
	/* wait DATA0 == 1 */
	while( !(inl( SDM_INFO2 ) & SDM_INFO2_DATA0_BIT) ) {
	}

	/* CMD9 : SEND_CSD */
	outl( 0, SDM_ARG0 );
	outl( 1, SDM_ARG1 );
	ret = emmc_send_cmd( 9 );
	if( ret != 0 ){
		return ret;
	}
	/* get erase size [block] */
	tmp_u16 = inh( SDM_RSP1 );
	block_num = (tmp_u16 & 0xE000) >> 13;
	tmp_u16 = inh( SDM_RSP2 );
	block_num |= (tmp_u16 & 0x0003) << 3;
	block_num ++;
	tmp_u16 = ((tmp_u16 & 0x007C) >> 2) + 1;
	block_num *= tmp_u16;
	emmc_erase_blocks = block_num;

	/* CMD7 : SELECT */
	outl( 0, SDM_ARG0 );
	outl( 1, SDM_ARG1 );
	ret = emmc_send_cmd( 7 );
	if( ret != 0 ){
		return ret;
	}

	for( i=0; i<(CFG_EMMC_PROTECT_SIZE>>2); i++ ) {
		*(wp_tbl++) = 0x01010101;
	}
	
	if( (CFG_ENV_EMMC_OFFSET + (CFG_ENV_SIZE/SDM_BLOCKLEN_VAL)) <= mbr_data->partitionTable[0].numberOfSectors ) {
		return 	(mbr_data->partitionTable[0].firstSectorNumbers + CFG_ENV_EMMC_OFFSET);
	} else {
		return 0;
	}
}

/* print */
int 
emmc_print_bank( unsigned long bank )
{
	int				ret;
	unsigned long	s_blk, e_blk;
	unsigned long	sect_size, sect_num;
	
	if( bank > CFG_EMMC_BANKS ) {
		return MP200_EMMC_ERR_PARAM1;
	}
	
	if( bank ) {
		printf("  Bank #%d: \n", bank);
		s_blk = (bank - 1) * CFG_EMMC_BLOCKS;
		e_blk = bank * CFG_EMMC_BLOCKS;
	} else {
		printf("  All : \n");
		s_blk = 0;
		e_blk = CFG_EMMC_BANKS * CFG_EMMC_BLOCKS;
	}
	
	sect_size = emmc_erase_blocks;
	sect_num = CFG_EMMC_BLOCKS / emmc_erase_blocks;
	printf("  Erase Size:   ");
	if( sect_size < 0x400 ) {
		printf(" %x Blocks in %d Erase Unit\n", sect_size, sect_num);
	} else if( sect_size < 0x100000 ) {
		printf("%xK Blocks in %d Erase Unit\n", (sect_size >> 10), sect_num);
	} else {
		printf("%xM Blocks in %d Erase Unit\n", (sect_size >> 20), sect_num);
	}

	sect_size = CFG_EMMC_SECT_OF_BLOCKS;
	sect_num = CFG_EMMC_SECTS;
	printf("  Protect Size: ");
	if( sect_size < 0x400 ) {
		printf(" %x Blocks in %d Sectors\n", sect_size, sect_num);
	} else if( sect_size < 0x100000 ) {
		printf("%xK Blocks in %d Sectors\n", (sect_size >> 10), sect_num);
	} else {
		printf("%xM Blocks in %d Sectors\n", (sect_size >> 20), sect_num);
	}

	printf("  Sector Start Block Number:\n    ");
	ret = emmc_check_protect( s_blk, e_blk, 1 );
	
	return ret;
}

/* erase */
int 
emmc_erase_sect( unsigned long s_blk, unsigned long e_blk )
{
	int					ret;
	unsigned long		chk_sblk, chk_eblk, size;
	struct boot_info	info;
	
	printf("erase check:   start = %x, end = %x\n", s_blk, e_blk);
	if( (s_blk & (emmc_erase_blocks - 1)) || (s_blk >= CFG_EMMC_BLOCKS) ) {
		return MP200_EMMC_ERR_PARAM1;
	}
	if( (e_blk & (emmc_erase_blocks - 1)) || (s_blk >= e_blk) || (e_blk > CFG_EMMC_BLOCKS) ) {
		return MP200_EMMC_ERR_PARAM2;
	}
	
	chk_sblk = (s_blk / CFG_EMMC_SECT_OF_BLOCKS) * CFG_EMMC_SECT_OF_BLOCKS;
	chk_eblk = ((e_blk + CFG_EMMC_SECT_OF_BLOCKS - 1) / CFG_EMMC_SECT_OF_BLOCKS) * CFG_EMMC_SECT_OF_BLOCKS;
	printf("protect check: start = %x, end = %x\n", chk_sblk, chk_eblk);
	ret = emmc_check_protect( chk_sblk, chk_eblk, 0 );
	if( ret ) {
		return MP200_EMMC_ERR_PROTECT;
	}
	
	size = e_blk - s_blk;
	info.erase = 1;							/* erase flag */
	while( size > 0 ) {
		/* set emmc erase information */
		info.block = s_blk;					/* erase block */
		if( size > MP200_ERASE_MAXBLK ) {
			info.size = MP200_ERASE_MAXBYTE;		/* erase size [byte] */
			size -= MP200_ERASE_MAXBLK;
			s_blk += MP200_ERASE_MAXBLK;
		} else {
			info.size = size * SDM_BLOCKLEN_VAL;	/* erase size [byte] */
			size = 0;
		}
		ret = emmc_erase( &info );
		if( ret != 0 ) {
			break;
		}
	}
	
	return ret;
}

int 
emmc_erase_bank( unsigned long bank )
{
	int				ret;
	unsigned long	s_blk, e_blk;

	if( bank > CFG_EMMC_BANKS ) {
		return MP200_EMMC_ERR_PARAM1;
	}
	
	if( bank ) {
		s_blk = (bank - 1) * CFG_EMMC_BLOCKS;
		e_blk = bank * CFG_EMMC_BLOCKS;
	} else {
		s_blk = 0;
		e_blk = CFG_EMMC_BANKS * CFG_EMMC_BLOCKS;
	}
	ret = emmc_erase_sect( s_blk, e_blk );
	
	return ret;
}

/* write */
int 
emmc_write_blk( unsigned long addr, unsigned long blk, unsigned int size)
{
	int					ret;
	unsigned long		chk_sblk, chk_eblk;
	struct boot_info	info;
	
	if( addr < PHYS_SDRAM_1 ) {
		return MP200_EMMC_ERR_PARAM1;
	}
	if( (addr + size) > (PHYS_SDRAM_2 + PHYS_SDRAM_2_SIZE) ) {
		return MP200_EMMC_ERR_PARAM3;
	}
	
	chk_sblk  = (blk / CFG_EMMC_SECT_OF_BLOCKS) * CFG_EMMC_SECT_OF_BLOCKS;
	chk_eblk  = ((size + CFG_EMMC_SECT_OF_BLOCKS - 1) / CFG_EMMC_SECT_OF_BLOCKS) * CFG_EMMC_SECT_OF_BLOCKS;
	chk_eblk += chk_sblk;
	printf("protect check: start = %x, end = %x\n", chk_sblk, chk_eblk);
	ret = emmc_check_protect( chk_sblk, chk_eblk, 0 );
	if( ret ) {
		return MP200_EMMC_ERR_PROTECT;
	}
	
	/* set emmc erase information */
	info.addr		= addr;	/* src address */
	info.block		= blk;	/* dst block */
	info.size		= size;	/* write size [byte] */
	ret = emmc_write( &info );
	if( ret ) {
		return MP200_EMMC_ERR_WRITE;
	}
	
	return ret;
}

int 
emmc_write_part( unsigned long addr, unsigned long part, unsigned int size)
{
	int				ret=0;
	unsigned long	blk, p_size;
	MBRecord		*mbr_data = (MBRecord *)ROMBOOT_DATA_START;
	
	blk = mbr_data->partitionTable[part].firstSectorNumbers;
	if( !blk ) {
		return MP200_EMMC_ERR_PARAM2;
	}
	
	p_size = mbr_data->partitionTable[part].numberOfSectors;
	if( ((size + SDM_BLOCKLEN_VAL - 1) / SDM_BLOCKLEN_VAL) > p_size ) {
		return MP200_EMMC_ERR_PARAM3;
	}
	ret = emmc_write_blk( addr, blk, size );
	
	return ret;
}

/* protect */
int 
emmc_protect_sect( unsigned long s_blk, unsigned long e_blk, int mode )
{
	unsigned char	*wp_tbl = (unsigned char *)CFG_EMMC_PROTECT_TBL;
	unsigned long	sect, e_sect;
	
	if( (s_blk & (CFG_EMMC_SECT_OF_BLOCKS - 1)) || (s_blk >= CFG_EMMC_BLOCKS) ) {
		return MP200_EMMC_ERR_PARAM1;
	}
	if( (e_blk & (CFG_EMMC_SECT_OF_BLOCKS - 1)) || (s_blk >= e_blk) || (e_blk > CFG_EMMC_BLOCKS) ) {
		return MP200_EMMC_ERR_PARAM2;
	}
	if( mode & 0xFFFFFFFE ) {
		return MP200_EMMC_ERR_PARAM3;
	}
	
	sect = s_blk / CFG_EMMC_SECT_OF_BLOCKS;
	wp_tbl += sect;
	e_sect = (e_blk / CFG_EMMC_SECT_OF_BLOCKS) - sect;
	if( !e_sect ) {
		e_sect = 1;
	}
	for( sect=0; sect<e_sect; sect++ ) {
		*(wp_tbl++) = mode;
	}

	return sect;
}

int 
emmc_protect_bank( unsigned long bank, int mode )
{
	int				ret;
	unsigned long	s_blk, e_blk;

	if( bank > CFG_EMMC_BANKS ) {
		return MP200_EMMC_ERR_PARAM1;
	}
	
	if( bank ) {
		s_blk = (bank - 1) * CFG_EMMC_BLOCKS;
		e_blk = bank * CFG_EMMC_BLOCKS;
	} else {
		s_blk = 0;
		e_blk = CFG_EMMC_BANKS * CFG_EMMC_BLOCKS;
	}
	ret = emmc_protect_sect( s_blk, e_blk, mode );
	
	return ret;
}

static int 
emmc_check_protect( unsigned long s_blk, unsigned long e_blk, int print )
{
	unsigned char	*wp_tbl = (unsigned char *)CFG_EMMC_PROTECT_TBL;
	unsigned long	sect, e_sect, p_blk;
	
	if( (s_blk & (CFG_EMMC_SECT_OF_BLOCKS - 1)) || (s_blk >= CFG_EMMC_BLOCKS) ) {
		return MP200_EMMC_ERR_PARAM1;
	}
	if( (e_blk & (CFG_EMMC_SECT_OF_BLOCKS - 1)) || (s_blk >= e_blk) || (e_blk > CFG_EMMC_BLOCKS) ) {
		return MP200_EMMC_ERR_PARAM2;
	}
	
	sect = s_blk / CFG_EMMC_SECT_OF_BLOCKS;
	wp_tbl += sect;
	e_sect = (e_blk / CFG_EMMC_SECT_OF_BLOCKS) - sect;
	if( print ) {
		p_blk = s_blk;
		for( sect=0; sect<e_sect; sect++ ) {
			printf("%08X ", p_blk);
			p_blk += CFG_EMMC_SECT_OF_BLOCKS;
			if( *(wp_tbl++) ) {
				printf("(RO) ");
			} else {
				printf("     ");
			}
			if( ((sect+1)%5) == 0 ) {
				printf("\n    ");
			}
		}
		printf("\n");
	} else {
		for( sect=0; sect<e_sect; sect++ ) {
			if( *(wp_tbl++) ) {
				return MP200_EMMC_ERR_PROTECT;
			}
		}
	}

	return MP200_EMMC_OK;
}

int 
emmc_erase( struct boot_info *boot_info )
{
	int				ret;
	unsigned int	src;

	src  = boot_info->block;
	if( !acc_mode ) {
		src  *= SDM_BLOCKLEN_VAL;
	}
	if( boot_info->erase ) {
		/* erase */
		ret = emmc_multi_block_erase( src, boot_info->size );
	} else {
		/* clear */
		ret = emmc_multi_block_write( src, (unsigned char *)NULL, boot_info->size );
	}
	
	return ret;
}

int 
emmc_write( struct boot_info *boot_info )
{
	int				ret;
	unsigned int	dest, src;

	dest = (unsigned int)boot_info->addr;
	src  = boot_info->block;
	if( !acc_mode ) {
		src  *= SDM_BLOCKLEN_VAL;
	}
	/* write */
	ret = emmc_multi_block_write( src, (unsigned char *)dest, boot_info->size );
	
	return ret;
}

int 
emmc_read( struct boot_info *boot_info )
{
	int				ret;
	unsigned int	dest, src;

	dest = (unsigned int)boot_info->addr;
	src  = boot_info->block;
	if( !acc_mode ) {
		src  *= SDM_BLOCKLEN_VAL;
	}
	/* read */
	ret = emmc_multi_block_read( src, (unsigned char *)dest, boot_info->size );
	
	return ret;
}
#endif	/* u-boot */

static int 
emmc_send_cmd( int cmd )
{
	int ret=0;

	outl( SDM_INFO_CLEAR, SDM_INFO1 );			/* Clear ALL */
	outl( SDM_INFO_CLEAR, SDM_INFO2 );			/* Clear ALL */
	
	cmd = cmd & 0x0000ffff;
	outl( cmd, SDM_CMD );
	while( !(inl( SDM_INFO1 ) & SDM_INFO1_RES_END) ) {
		if( inl( SDM_INFO2 ) & SDM_INFO2_ERR_ALL ) {
			break;
		}
	}
	ret = inl( SDM_INFO2 ) & SDM_INFO2_ERR_ALL;
	if( ret != 0 ) {
		outl( (inl( SDM_INFO2 ) & ~SDM_INFO2_ERR_ALL), SDM_INFO2 );
	} else {
		outl( (inl(SDM_INFO1) & ~SDM_INFO1_RES_END), SDM_INFO1 );
	}

	return ret;
}

static int
emmc_multi_block_read( unsigned int address, unsigned char *data, unsigned int size )
{
	int				ret, len, i, j;
	unsigned int	info2_tmp;
	unsigned short *ptr = (unsigned short *)data;
	unsigned int block_num;

	if( (data == NULL) || (size == 0) ) {
		return -1;
	}

	block_num = (size / SDM_BLOCKLEN_VAL);
	if( size & (SDM_BLOCKLEN_VAL-1) ) {
		block_num ++;
	}

	outl( block_num, SDM_SECCNT );
	outl( 0x100, SDM_STOP );

	/* CMD18 : READ_MULTI_BLOCK */
	outl( (address & 0xFFFF),         SDM_ARG0 );
	outl( ((address >> 16) & 0xFFFF), SDM_ARG1 );
	ret = emmc_send_cmd( 18 );
	if( ret != 0 ){
		return ret;
	}

	len = inl( SDM_SIZE );
	for ( j = 0; j < block_num; j++ ) {
		while( 1 ) {
			/* wait read enable */
			info2_tmp = inl( SDM_INFO2 );
			if( info2_tmp & (SDM_INFO2_BRE_BIT | SDM_INFO2_ERR_ALL) ) {
				break;
			}
		}
		if( info2_tmp & SDM_INFO2_ERR_ALL ) {
			outl( (info2_tmp & ~SDM_INFO2_ERR_ALL), SDM_INFO2 );
			return -1;
		}

		outl( (info2_tmp & ~SDM_INFO2_BRE_BIT), SDM_INFO2 );
		/* 1block read */
		for( i=0; i<(len/2); i++ ) {
			*ptr++ = inl( SDM_BUF0 );
		}
	}

	/* wait for data read end */
	while( !(inl( SDM_INFO1 ) & SDM_INFO1_RES_RW_END) ) {
		if( inl( SDM_INFO2 ) & SDM_INFO2_ERR_ALL ){
			outl( (inl( SDM_INFO2 ) & ~SDM_INFO2_ERR_ALL), SDM_INFO2 );
			return -1;
		}
	}
	outl( (inl( SDM_INFO1 ) & ~(SDM_INFO1_RES_END | SDM_INFO1_RES_RW_END)), SDM_INFO1 );

	return 0;
}

#if !defined(MP200_MINIBOOT)
static int
emmc_multi_block_write( unsigned int address, unsigned char *data, unsigned int size )
{
	int				ret, len, i, j;
	unsigned int	info2_tmp;
	unsigned short *ptr = (unsigned short *)data;
	unsigned int	block_num;

	if( size == 0 ) {
		return -1;
	}

	block_num = (size / SDM_BLOCKLEN_VAL);
	if( size & (SDM_BLOCKLEN_VAL-1) ) {
		block_num ++;
	}

	outl( block_num, SDM_SECCNT );
	outl( 0x100, SDM_STOP );

	/* CMD25 : WRITE_MULTI_BLOCK */
	outl( (address & 0xFFFF),         SDM_ARG0 );
	outl( ((address >> 16) & 0xFFFF), SDM_ARG1 );
	ret = emmc_send_cmd( 25 );
	if( ret != 0 ){
		return ret;
	}

	len = inl( SDM_SIZE );
	for ( j = 0; j < block_num; j++ ) {
		while( 1 ) {
			/* wait write enable */
			info2_tmp = inl( SDM_INFO2 );
			if( info2_tmp & (SDM_INFO2_BWE_BIT | SDM_INFO2_ERR_ALL) ) {
				break;
			}
		}
		if( info2_tmp & SDM_INFO2_ERR_ALL ) {
			outl( (info2_tmp & ~SDM_INFO2_ERR_ALL), SDM_INFO2 );
			return -1;
		}

		outl( (info2_tmp & ~SDM_INFO2_BWE_BIT), SDM_INFO2 );
		if( data != NULL ) {
			/* 1block write */
			for( i=0; i<(len/2); i++ ) {
				outl( *ptr++, SDM_BUF0 );
			}
		} else {
			/* 1block clear */
			for( i=0; i<(len/2); i++ ) {
				outl( 0, SDM_BUF0 );
			}
		}
	}

	/* wait for data read end */
	while( !(inl( SDM_INFO1 ) & SDM_INFO1_RES_RW_END) ) {
		if( inl( SDM_INFO2 ) & SDM_INFO2_ERR_ALL ){
			outl( (inl( SDM_INFO2 ) & ~SDM_INFO2_ERR_ALL), SDM_INFO2 );
			return -1;
		}
	}
	outl( (inl( SDM_INFO1 ) & ~(SDM_INFO1_RES_END | SDM_INFO1_RES_RW_END)), SDM_INFO1 );

	return 0;
}


static int
emmc_multi_block_erase( unsigned int address, unsigned int size )
{
	int				ret;
	unsigned int	end_block;

	printf("eMMC Erase: start %08x[block] length %08x[block]\n", address, (size / SDM_BLOCKLEN_VAL));
	if( size == 0 ) {
		return -1;
	}

	end_block = address + (size / SDM_BLOCKLEN_VAL) - 1;
	if( size & (SDM_BLOCKLEN_VAL-1) ) {
		end_block ++;
	}

	/* CMD35 : ERASE_GROUP_START */
	outl( (address & 0xFFFF),         SDM_ARG0 );
	outl( ((address >> 16) & 0xFFFF), SDM_ARG1 );
	ret = emmc_send_cmd( 35 );
	if( ret != 0 ){
		return ret;
	}

	/* CMD36 : ERASE_GROUP_END */
	outl( (end_block & 0xFFFF),         SDM_ARG0 );
	outl( ((end_block >> 16) & 0xFFFF), SDM_ARG1 );
	ret = emmc_send_cmd( 36 );
	if( ret != 0 ){
		return ret;
	}

	/* CMD38 : ERASE */
	outl( 0, SDM_ARG0 );
	outl( 0, SDM_ARG1 );
	ret = emmc_send_cmd( 38 );
	if( ret != 0 ){
		return ret;
	}
	/* wait DATA0 == 1 */
	while( !(inl( SDM_INFO2 ) & SDM_INFO2_DATA0_BIT) ) {
	}

	return 0;
}

#else	/* defined(MP200_MINIBOOT) */
#define IMAGE_DATA_SIZE		0x0c

void 
mini_boot( void )
{
	int				ret;
	int				*acc_mode = (int *)SRAM_ACC_MODE;
	unsigned int	dest, src, k_size;
	MBRecord		*mbr_data = (MBRecord *)ROMBOOT_DATA_START;
#ifndef MP200_EMMC_ACCMODE
	unsigned int	first_sector;
#endif
	unsigned int	last_sector;
	unsigned char	*head_ptr;
	

	uart_init();

	last_sector = (inl(SDM_ARG0) & 0xffff) | ((inl(SDM_ARG1) & 0xffff)<< 16);
#ifdef MP200_EMMC_ACCMODE
	*acc_mode = MP200_EMMC_ACCMODE;
#else	/* !defined(MP200_EMMC_ACCMODE) */
	/* check access mode */
	first_sector  = mbr_data->partitionTable[0].firstSectorNumbers;
	if( (last_sector - first_sector) < SDM_BLOCKLEN_VAL ) {
		*acc_mode = 1;
	} else {
		*acc_mode = 0;
	}
#endif	/* !defined(MP200_EMMC_ACCMODE) */

	/* change high speed mode */
	outl( (inl( SDM_OPTION ) | 0x00ee), SDM_OPTION );
	/* CMD6 : SWITCH (set HS_TIMING on) */
    outl( (0x0001                      << 8),                      SDM_ARG0 );
    outl( ((MMC_SWITCH_MODE_WRITE_BYTE << 8) | EXT_CSD_HS_TIMING), SDM_ARG1 );
    emmc_send_cmd( 0x0546 );
	/* wait DATA0 == 1 */
	while( !(inl( SDM_INFO2 ) & SDM_INFO2_DATA0_BIT) ) {
	}
	/* Clken and SystemCLK Div2(set 41.65MHz) */
	outl( SDM_CLK_CTRL_DIV2, SDM_CLK_CTRL );

	dest = (unsigned int)U_BOOT_TEXTADDR;
	if( *acc_mode ) {
		src  = last_sector + 1;
	} else {
		src  = last_sector + SDM_BLOCKLEN_VAL;
	}
	/* read */
	ret = emmc_multi_block_read( src, (unsigned char *)dest, U_BOOT_SIZE );
	if( ret != 0 ) {
		dbg_print("read error\n");
		goto ERR_END;
	}

	dest = (unsigned int)KERNEL_START;
	if( *acc_mode ) {
		src  = mbr_data->partitionTable[1].firstSectorNumbers;
	} else {
		src  = mbr_data->partitionTable[1].firstSectorNumbers * SDM_BLOCKLEN_VAL;
	}
	ret = emmc_multi_block_read( src, (unsigned char *)dest, SDM_BLOCKLEN_VAL );
	if( ret != 0 ) {
		dbg_print("read error\n");
		goto ERR_END;
	}
	head_ptr = (unsigned char *)dest + IMAGE_DATA_SIZE;
	k_size = (unsigned int)(*head_ptr++);
	k_size <<= 8;
	k_size |= (unsigned int)(*head_ptr++);
	k_size <<= 8;
	k_size |= (unsigned int)(*head_ptr++);
	k_size <<= 8;
	k_size |= (unsigned int)*head_ptr;
	ret = emmc_multi_block_read( src, (unsigned char *)dest, (k_size + 64) );
	if( ret != 0 ) {
		dbg_print("read error\n");
		goto ERR_END;
	}

	/* jump mini-boot */
	((void (*)(void))U_BOOT_TEXTADDR)();

 ERR_END:
	outl( 0x0020, SDM_CLK_CTRL ); /* Clk Disable */

	return;
}
#endif	/* defined(MP200_MINIBOOT) */

#endif	/* defined(CONFIG_MP200_EMMCBOOT) */

