/*******************************************************************************
* Copyright (C) NEC Electronics Corporation 2008
* NEC ELECTRONICS CONFIDENTIAL AND PROPRIETARY
* All rights reserved by NEC Electronics Corporation.
* Use of copyright notice does not evidence publication
*
*  File Name : emmc/emmc.h
*  Contents  : emmc header
*  History   :
*   Date        Rev.      Who         Details     Keyword
*  ==========  ========  ==========  ==========  ==========
*  2009/8/19   Rev1.01   j-nakamu
*
*******************************************************************************/

/* register address define */
/* base address */
#define	EM1_SD2_BASE			0x50090000

#define SDM_CMD					(EM1_SD2_BASE + 0x0000)
#define SDM_PORTSEL				(EM1_SD2_BASE + 0x0004)
#define SDM_ARG0				(EM1_SD2_BASE + 0x0008)
#define SDM_ARG1				(EM1_SD2_BASE + 0x000c)
#define SDM_STOP				(EM1_SD2_BASE + 0x0010)
#define SDM_SECCNT				(EM1_SD2_BASE + 0x0014)
#define SDM_RSP0				(EM1_SD2_BASE + 0x0018)
#define SDM_RSP1				(EM1_SD2_BASE + 0x001c)
#define SDM_RSP2				(EM1_SD2_BASE + 0x0020)
#define SDM_RSP3				(EM1_SD2_BASE + 0x0024)
#define SDM_RSP4				(EM1_SD2_BASE + 0x0028)
#define SDM_RSP5				(EM1_SD2_BASE + 0x002c)
#define SDM_RSP6				(EM1_SD2_BASE + 0x0030)
#define SDM_RSP7				(EM1_SD2_BASE + 0x0034)

#define SDM_INFO1				(EM1_SD2_BASE + 0x0038)
#define SDM_INFO2				(EM1_SD2_BASE + 0x003c)
#define SDM_INFO1_MASK			(EM1_SD2_BASE + 0x0040)
#define SDM_INFO2_MASK			(EM1_SD2_BASE + 0x0044)
#define SDM_CLK_CTRL			(EM1_SD2_BASE + 0x0048)
#define SDM_SIZE				(EM1_SD2_BASE + 0x004C)

#define SDM_OPTION				(EM1_SD2_BASE + 0x0050)

#define SDM_BUF0				(EM1_SD2_BASE + 0x0060)

#define SDM_SOFT_RST			(EM1_SD2_BASE + 0x01c0)

/* register set value define */
#define SDM_SOFT_RST_MDL		0x00000000	/* module reset */
#define SDM_SOFT_RST_SFT		0x00000007	/* release reset */
#define SDM_OPTION_VAL			0x00008088	/* response time out counter */
#define SDM_INFO_CLEAR			0x00000000	/* interrupt status clear */
#define SDM_INFO_MASK_ALL		0x0000ffff	/* interrupt mask */
#define SDM_INFO_MASK_FREE		0x00000000	/* interrupt enable */
#define SDM_CLK_CTRL_DIV256		0x00000140	/* enable SD clock(LBUS(=57.344) * 1/256) no control */
#define SDM_CLK_CTRL_DIV4		0x00000301	/* enable SD clock(LBUS(=57.344) * 1/  4) auto control */
#define SDM_CLK_CTRL_DIV2		0x00000700	/* enable SD clock(LBUS(=83.3) * 1/  2) auto control */
#define SDM_INFO1_MASK_VAL		0x0000fffa	/* Response end,R/W access all end Enable */
#define SDM_INFO2_MASK_VAL		0x0000807F	/*  */

#define SDM_INFO1_RES_END		0x0001		/* RES END BIT */
#define SDM_INFO1_RES_RW_END	0x0004      /* RES Read/Write END BIT */

#define SDM_INFO2_ERR_ALL		0x807F		/* INFO2 ERR */
#define SDM_INFO2_BWE_BIT		0x0200      /* INFO2 BWE BIT */
#define SDM_INFO2_BRE_BIT		0x0100		/* INFO2 BRE BIT */
#define SDM_INFO2_DATA0_BIT		0x0080		/* INFO2 DATA0 BIT */
#define SDM_INFO2_RES_TOUT		0x0040		/* INFO2 ERR6 BIT */

#define SDM_BLOCKLEN_VAL		512
#define SDM_HALF_BLOCKLEN_VAL	256

#define SDM_OPT_WIDTH_BIT		0x8000

#define SDM_STOP_SEC_BIT		0x0100


#define EMMC_MINI_BOOT_SIZE		4096		/* 4k */

#define MBR_ADDR				0			/* 0x0000-0x01FF */
#define MBR_PTABLE_NUM			4			/* Number of Partition Table in MBR */
#define MBR_ACTIVE_PT			0x80		/* Active Partition Table */

#define KERNEL_START			CFG_LOAD_ADDR	/* SDRAM */
#define KERNEL_SIZE				0x380000		/* 3.5M */


/*
 * MMC_SWITCH argument format:
 *
 *	[31:26] Always 0
 *	[25:24] Access Mode
 *	[23:16] Location of target Byte in EXT_CSD
 *	[15:08] Value Byte
 *	[07:03] Always 0
 *	[02:00] Command Set
 */

/*
 * EXT_CSD fields
 */

#define EXT_CSD_BUS_WIDTH           183     /* WO */
#define EXT_CSD_HS_TIMING           185     /* R/W */
#define EXT_CSD_CARD_TYPE           196     /* RO */
#define EXT_CSD_SEC_CNT             212     /* RO, 4 bytes */

/*
 * EXT_CSD field definitions
 */

#define EXT_CSD_CMD_SET_NORMAL      0x01
#define EXT_CSD_CMD_SET_SECURE      0x02
#define EXT_CSD_CMD_SET_CPSECURE    0x04

#define EXT_CSD_CARD_TYPE_26        0x01    /* Card can run at 26MHz */
#define EXT_CSD_CARD_TYPE_52        0x02    /* Card can run at 52MHz */

#define EXT_CSD_BUS_WIDTH_1         0x00    /* Card is in 1 bit mode */
#define EXT_CSD_BUS_WIDTH_4         0x01    /* Card is in 4 bit mode */
#define EXT_CSD_BUS_WIDTH_8         0x02    /* Card is in 8 bit mode */

/*
 * MMC_SWITCH access modes
 */

#define MMC_SWITCH_MODE_CMD_SET     0x00    /* Change the command set */
#define MMC_SWITCH_MODE_SET_BITS    0x01    /* Set bits which are 1 in value */
#define MMC_SWITCH_MODE_CLEAR_BITS  0x02    /* Clear bits which are 1 in value */
#define MMC_SWITCH_MODE_WRITE_BYTE  0x03    /* Set target to value */


typedef struct MasterBootRecord {
	unsigned char nondata[446];
	struct {
		unsigned char	bootDescriptor;
		unsigned char	firstPartitionSector[3];
		unsigned char	fileSystemDescriptor;
		unsigned char	lastPartitionSector[3];
		unsigned int	firstSectorNumbers;
		unsigned int	numberOfSectors;
	} partitionTable[4];
	unsigned char    sig[2]; 
} __attribute__ ((packed)) MBRecord;

