/*******************************************************************************
* Copyright (C) NEC Electronics Corporation 2008
* NEC ELECTRONICS CONFIDENTIAL AND PROPRIETARY
* All rights reserved by NEC Electronics Corporation.
* Use of copyright notice does not evidence publication
*
*  File Name : include/config.h
*  Contents  : Mini boot configration header
*  History   :
*   Date        Rev.      Who         Details     Keyword
*  ==========  ========  ==========  ==========  ==========
*  2009/8/19   Rev1.01   j-nakamu
*
*******************************************************************************/

#ifndef __MINI_CONFIG_H
#define __MINI_CONFIG_H

#ifndef NULL
#define NULL	0
#endif

#define inl(a)						(*(volatile unsigned long *)(a))
#define outl(v,a)					(*(volatile unsigned long *)(a) = (v))
#define inh(a)						(*(volatile unsigned short *)(a))
#define outh(v,a)					(*(volatile unsigned short *)(a) = (v))

/* SRAM(128K 0xa0000000-0xa001ffff)
 *	0xa0000000 - 0xa0010000	mini-boot TEXT & DATA area
 * 
 *	0xa0019000 - ?			rom-boot DATA area
 *	?          - 0xa001feff	STACK area
 * 
 *	0xa001fff0				FATAL ERR VECTOR
 *	0xa001fff4				FIQ INT VECTOR
 *	0xa001fff8				IRQ INT VECTOR
 *	0xa001fffC				SWI INT VECTOR
 */
#define EM1_SRAM_START				0xa0000000
#define EM1_SRAM_END				0xa0020000

#define MINIBOOT_OFFSET				0x00000000
#define ROMBOOT_DATA_OFFSET			0x00019000

#define ROMBOOT_DATA_START			(EM1_SRAM_START + ROMBOOT_DATA_OFFSET)
#define SRAM_ACC_MODE				(ROMBOOT_DATA_START + 0x200)

#define STACK_START					(EM1_SRAM_END - 0x104)
#define SUB_IRQ_STACK				0x0
#define SUB_SVC_STACK				0x200

#define SRAM_SWI_VECTOR				(EM1_SRAM_END - 0x04)
#define SRAM_IRQ_VECTOR				(EM1_SRAM_END - 0x08)
#define SRAM_FIQ_VECTOR				(EM1_SRAM_END - 0x0c)
#define FATAL_ERR_VECTOR			(EM1_SRAM_END - 0x10)

#define CORRECT_DATA_SAVE_ADDR		(MINIBOOT_START + MINIBOOT_SIZE)

/* NAND
 *	0x00000000 - 0x000007ff	miniboot(master)
 *	0x00000800 - 0x00000fff	miniboot(mirror)
 */
#define MINIBOOT_NAND_START			0
#define MINIBOOT_NAND_SIZE			MINIBOOT_SIZE

#define MINIBOOT_NAND_MIRROR_START	(MINIBOOT_NAND_START + MINIBOOT_NAND_SIZE)
#define MINIBOOT_NAND_MIRROR_SIZE	MINIBOOT_NAND_SIZE

#endif /* __MINI_CONFIG_H */
