/*******************************************************************************
* Copyright (C) NEC Electronics Corporation 2008
* NEC ELECTRONICS CONFIDENTIAL AND PROPRIETARY
* All rights reserved by NEC Electronics Corporation.
* Use of copyright notice does not evidence publication
*
*  File Name : main/etc.c
*  Contents  : shared functions
*  History   :
*   Date        Rev.      Who         Details     Keyword
*  ==========  ========  ==========  ==========  ==========
*  2009/8/19   Rev1.01   j-nakamu
*
*******************************************************************************/

#include "config.h"
#include "mini_config.h"
#include "serial.h"

#define UART_1

/* SMU */
#define SMU_DIVU70SCLK				(0xc0110158)
#define	SMU_DIVU70SCLK_VAL			0x00000030	/* 229.376MHz(=PLL3(229.376MHz) * 1/ 4) */
#define SMU_DIVU71SCLK				(0xc011015c)
#define	SMU_DIVU70SCLK_VAL			0x00000030	/* 229.376MHz(=PLL3(229.376MHz) * 1/ 4) */

#if defined(CONFIG_MP200_SDBOOT)
#define MINI_DEBUG
#endif

#ifdef MINI_DEBUG
static void hex2char(unsigned long val);
#endif

int
memcmp(const void *src1, const void *src2, int count)
{
	const unsigned char *src1_tmp = (const unsigned char *)src1;
	const unsigned char *src2_tmp = (const unsigned char *)src2;
	int ret = 0;

	while (count) {
		if ((ret = *src1_tmp++ - *src2_tmp++) != 0)
			break;
		count--;
	}

	return ret;
}

/* uart initial Function */
void
uart_init(void)
{
#ifdef UART_1
	/* UART1 Setting */
	/* default : SCLK=229.376M, 115200bps, 8bit, nonparity */
	outl(SMU_DIVU71SCLK_VAL, SMU_DIVU71SCLK);

	outh(UART_IER_CLR, UART1_IER );
	outh((UART_LCR_DLAB | UART_LCR_WLEN8), UART1_LCR );
	outh(UART_DLL_115200, UART1_DLL );
	outh(UART_DLM_115200, UART1_DLM );
	outh(UART_LCR_WLEN8, UART1_LCR );
	outh((UART_MCR_RTS | UART_MCR_DTR), UART1_MCR );
	outh(UART_FCR_FIFO_ENABLE, UART1_FCR );
	outh(UART_FCR_FIFO_RESET, UART1_FCR );
#else
	/* UART0 Setting */
	/* default : SCLK=229.376M, 115200bps, 8bit, nonparity */
	outl(SMU_DIVU70SCLK_VAL, SMU_DIVU70SCLK);

	outh(UART_IER_CLR, UART0_IER );
	outh((UART_LCR_DLAB | UART_LCR_WLEN8), UART0_LCR );
	outh(UART_DLL_115200, UART0_DLL );
	outh(UART_DLM_115200, UART0_DLM );
	outh(UART_LCR_WLEN8, UART0_LCR );
	outh((UART_MCR_RTS | UART_MCR_DTR), UART0_MCR );
	outh(UART_FCR_FIFO_ENABLE, UART0_FCR );
	outh(UART_FCR_FIFO_RESET, UART0_FCR );
#endif
}

/* send character Function */
void
putc(unsigned char c)
{
#ifdef MINI_DEBUG
#ifdef UART_1
	if (c == '\n') {
		/* Wait until there is space in the FIFO */
		while ((inh(UART1_LSR) & UART_LSR_THRE) == 0) {
			;
		}
		outh('\r', UART1_THR);
	}

	/* Wait until there is space in the FIFO */
	while ((inh(UART1_LSR) & UART_LSR_THRE) == 0) {
		;
	}
	/* Send the character */
	outh(c, UART1_THR);
#else
	if (c == '\n') {
		/* Wait until there is space in the FIFO */
		while ((inh(UART0_LSR) & UART_LSR_THRE) == 0) {
			;
		}
		outh('\r', UART0_THR);
	}

	/* Wait until there is space in the FIFO */
	while ((inh(UART0_LSR) & UART_LSR_THRE) == 0) {
		;
	}
	/* Send the character */
	outh(c, UART0_THR);
#endif
#endif
}

/* send stream Function */
void
dbg_print(const char* s)
{
#ifdef MINI_DEBUG
	for ( ; *s != '\0'; s++) {
		if (*s == '\n') {
			putc('\r');
			putc('\n');
		} else {
			putc(*s);
		}
	}
#endif
}

void
dbg_print_val(const char* s, unsigned long num)
{
#ifdef MINI_DEBUG
	for ( ; *s != '\0'; s++) {
		if (*s == '\n') {
			putc('\r');
			putc('\n');
		} else if (*s == '%') {
			s++;
			if (*s == 'c') {
				putc(*(char *)num);
			} else if (*s == 's') {
				dbg_print((char *)num);
			} else if (*s == 'x') {
				hex2char(num);
			} else {
				putc('%');
				putc(*s);
			}
		} else {
			putc(*s);
		}
    }
#endif
}

#ifdef MINI_DEBUG
static void
hex2char( unsigned long val )
{
	char			buf[10];
	char			*tmp_ptr = buf;
	int				i=1;
	unsigned int	tmp, uint16=val;
	
	while( 1 ) {
		uint16 >>= 4;
		if( uint16 == 0 ) {
			break;
		}
		i++;
	}
	*(tmp_ptr + i) = '\0';
	
	tmp_ptr += (i - 1);
	uint16 = val;
	while( i ) {
		tmp = uint16 & 0x0000000F;
		if( tmp < 10 ) {
			*tmp_ptr-- = (char)('0' + tmp);
		} else {
			*tmp_ptr-- = (char)('a' + tmp - 10);
		}
		uint16 >>= 4;
		i--;
	}
	dbg_print(buf);
}
#endif
