/*******************************************************************************
* Copyright (C) NEC Electronics Corporation 2008
* NEC ELECTRONICS CONFIDENTIAL AND PROPRIETARY
* All rights reserved by NEC Electronics Corporation.
* Use of copyright notice does not evidence publication
*
*  File Name : sd/sd.c
*  Contents  : SD boot main
*  History   :
*   Date        Rev.      Who         Details     Keyword
*  ==========  ========  ==========  ==========  ==========
*  2009/8/19   Rev1.01   j-nakamu
*
*******************************************************************************/

#include <config.h>

#if	defined(CONFIG_MP200_SDBOOT)

#include "../include/mini_config.h"
#include "sd.h"

/* static function (mini-boot & u-boot) */
static int sd_send_cmd( int cmd );
static int sd_multi_block_read( unsigned int address, unsigned char *data, unsigned int size);
static int sd_mbr_pram_chk(MBRecord *mbr);
static void sd_read_data(unsigned char * buf);
static int sd_1sector_read(int tSector);

#if defined(MP200_MINIBOOT)	
/* global function */
void mini_boot( void );

/* extern function */
extern void uart_init(void);
extern void dbg_print(const char* s);
extern int memcmp(const void*, const void*, int);
#else	/* u-boot */
#define MP200_SD_OK				0
#define MP200_SD_ERR_PARAM1		-1
#define MP200_SD_ERR_PARAM2		-2
#define MP200_SD_ERR_PARAM3		-3
#define MP200_SD_ERR_PROTECT	-4
#define MP200_SD_ERR_WRITE		-5
#define MP200_SD_ERR_READ		-6

/* boot information */
struct boot_info {
	unsigned long addr;		/* ddr address */
	unsigned long block;	/* block number */
	unsigned long size;		/* size [byte] */
	unsigned long erase;	/* erase flag */
};

/* static function definition */
static int sd_check_protect( unsigned long, unsigned long, int );
static int sd_multi_block_write( unsigned int address, unsigned char *data, unsigned int size);
static int sd_multi_block_erase( unsigned int address, unsigned int size );

/* global variable */
int sd_erase_blocks = 0;

/* static variable */
static int acc_mode		= 0;
static int sd_ver		= 0;
static int dev_total_blk = 0;
static int read_blk_len	= 0;
static int dev_size		= 0;

unsigned long 
sd_flash_init( void )
{
	int				i, ret, block_num, c_size_n;
	unsigned short	tmp_u16;
	unsigned int	RCA_VAL;
	unsigned int	*wp_tbl = (unsigned int *)CFG_SD_PROTECT_TBL;
	MBRecord		*mbr_data;
	unsigned char	*read_data = (unsigned char *)ROMBOOT_DATA_START;
	int				c_size_mult[8] = { 4, 8, 16, 32, 64, 128, 256, 512};
	int				blk_len[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 512, 1024, 2048, 0, 0, 0, 0};

	mbr_data = (MBRecord *)read_data;
	ret = sd_1sector_read(0);
	sd_read_data(read_data);

	acc_mode = *(int *)SRAM_ACC_MODE;	/* 0xa0019200 */
	
	/* CMD7 : DESELECT_CARD */
	outl( 0, SDM_ARG0 );
	outl( 0, SDM_ARG1 );
	ret = sd_send_cmd( 7 );
	/* wait DATA0 == 1 */
	while( !(inl( SDM_INFO2 ) & SDM_INFO2_DATA0_BIT) ) {
	}

	/* CMD3 SET */
	outl(0, SDM_ARG0);
	outl(0, SDM_ARG1);

	ret = sd_send_cmd(3); /* CMD3 */
	if(ret != 0) {
		return ret;
	}
	RCA_VAL = inl(SDM_RSP1);

	/* CMD9 : SEND_CSD */
	outl( 0, SDM_ARG0 );
	outl( RCA_VAL, SDM_ARG1 );
	ret = sd_send_cmd( 9 );
	if(ret != 0) {
		return ret;
	}
	/* get SD Version  [Ver:1.0 or 2.0] */
	tmp_u16 = inh( SDM_RSP7 );
	sd_ver = (tmp_u16 & 0x00C0) >> 6;			/* 0:Ver1.0 1:Ver2.0 */

	/* get sd device size [block] */
	if(sd_ver == 0) {
		tmp_u16 = inh( SDM_RSP2 );				/* C_SIZE_MULTI */
		c_size_n = (tmp_u16 & 0x0380) >> 7;
		tmp_u16 = inh( SDM_RSP3 );				/* C_SIZE */
		dev_size = (tmp_u16 & 0xFFC0) >> 6;
		tmp_u16 = inh( SDM_RSP4 );				/* C_SIZE */
		dev_size |= (tmp_u16 & 0x0003) << 10;
		dev_size += 1;
		
		dev_size = dev_size * c_size_mult[c_size_n];
		
		tmp_u16 = inh( SDM_RSP4 );				/* READ_BL_LEN */
		read_blk_len = (tmp_u16 & 0x0F00) >> 8;
		read_blk_len = blk_len[read_blk_len];

		dev_total_blk = (dev_size * read_blk_len) / SDM_BLOCKLEN_VAL;

	} else if(sd_ver == 1) {
		tmp_u16 = inh( SDM_RSP2 );
		dev_size = (tmp_u16 & 0xFF00) >> 8;
		tmp_u16 = inh( SDM_RSP3 );
		dev_size |= (tmp_u16 & 0x3FFF) << 8;
		dev_size += 1;
				
		dev_total_blk = (dev_size * SDM_BLOCKLEN_VAL) / SDM_BLOCKLEN_VAL;
	}
	
	/* get erase size [block] */
	tmp_u16 = inh( SDM_RSP2 );					/* ERASE_BLK_EN */
	sd_erase_blocks = (tmp_u16 & 0x0040) >> 6;
	if(sd_erase_blocks == 0) {
		tmp_u16 = inh( SDM_RSP1 );				/* SECTOR_SIZE */
		block_num = (tmp_u16 & 0x8000) >> 15;
		tmp_u16 = inh( SDM_RSP2 );				/* SECTOR_SIZE */
		block_num |= (tmp_u16 & 0x003F) << 1;
		sd_erase_blocks = block_num;
	} else {
		sd_erase_blocks = 1;
	}

	/* CMD7 : SELECT */
	outl( 0, SDM_ARG0 );
	outl( RCA_VAL, SDM_ARG1 );
	ret = sd_send_cmd( 7 );
	if( ret != 0 ){
		return ret;
	}
	for( i=0; i<((dev_total_blk/CFG_SD_SECT_OF_BLOCKS)>>2); i++ ) {
		*(wp_tbl++) = 0x01010101;
	}
	
	if( (CFG_ENV_SD_OFFSET + (dev_total_blk/SDM_BLOCKLEN_VAL)) <= mbr_data->partitionTable[1].numberOfSectors ) {
		return 	(mbr_data->partitionTable[1].firstSectorNumbers + CFG_ENV_SD_OFFSET);
	} else {
		return 0;
	}
}

/* print */
int 
sd_print_bank( unsigned long bank )
{
	int				ret;
	unsigned long	s_blk, e_blk;
	unsigned long	sect_size, sect_num;
	
	if( bank > CFG_SD_BANKS ) {
		return MP200_SD_ERR_PARAM1;
	}
	
	if( bank ) {
		printf("  Bank #%d: \n", bank);
		s_blk = (bank - 1) * dev_total_blk;
		e_blk = bank * dev_total_blk;
	} else {
		printf("  All : \n");
		s_blk = 0;
		e_blk = CFG_SD_BANKS * dev_total_blk;
	}
	
	sect_size = sd_erase_blocks;
	sect_num = dev_total_blk / sect_size;
	printf("  Erase Size:   ");
	if( sect_size < 0x400 ) {
		printf(" %x Blocks in %d Erase Unit\n", sect_size, sect_num);
	} else if( sect_size < 0x100000 ) {
		printf("%xK Blocks in %d Erase Unit\n", (sect_size >> 10), sect_num);
	} else {
		printf("%xM Blocks in %d Erase Unit\n", (sect_size >> 20), sect_num);
	}

	sect_size = CFG_SD_SECT_OF_BLOCKS;
	sect_num = ( dev_total_blk / sect_size );

	printf("  Protect Size: ");
	if( sect_size < 0x400 ) {
		printf(" %x Blocks in %d Sectors\n", sect_size, sect_num);
	} else if( sect_size < 0x100000 ) {
		printf("%xK Blocks in %d Sectors\n", (sect_size >> 10), sect_num);
	} else {
		printf("%xM Blocks in %d Sectors\n", (sect_size >> 20), sect_num);
	}

	printf("  Sector Start Block Number:\n    ");
	ret = sd_check_protect( s_blk, e_blk, 1 );
	
	return ret;
}

/* erase */
int 
sd_erase_sect( unsigned long s_blk, unsigned long e_blk )
{
	int					ret;
	unsigned long		chk_sblk, chk_eblk;
	struct boot_info	info;
	printf("erase check:   start = %x, end = %x\n", s_blk, e_blk);
	if( (s_blk & (sd_erase_blocks - 1)) || (s_blk >= dev_total_blk) ) {
		return MP200_SD_ERR_PARAM1;
	}
	if( (e_blk & (sd_erase_blocks - 1)) || (s_blk >= e_blk) || (e_blk > dev_total_blk) ) {
		return MP200_SD_ERR_PARAM2;
	}
	
	chk_sblk = (s_blk / CFG_SD_SECT_OF_BLOCKS) * CFG_SD_SECT_OF_BLOCKS;
	chk_eblk = ((e_blk + CFG_SD_SECT_OF_BLOCKS - 1) / CFG_SD_SECT_OF_BLOCKS) * CFG_SD_SECT_OF_BLOCKS;
	printf("protect check: start = %x, end = %x\n", chk_sblk, chk_eblk);
	ret = sd_check_protect( chk_sblk, chk_eblk, 0 );
	if( ret ) {
		return MP200_SD_ERR_PROTECT;
	}
	/* set sd erase information */
	info.block		= chk_sblk;									/* erase block */
	info.size		= (chk_eblk - chk_sblk) * SDM_BLOCKLEN_VAL;	/* erase size [byte] */
	info.erase		= 1;										/* erase flag */
	ret = sd_erase( &info );
	return ret;
}

int 
sd_erase_bank( unsigned long bank )
{
	int				ret;
	unsigned long	s_blk, e_blk;
	if( bank > CFG_SD_BANKS ) {
		return MP200_SD_ERR_PARAM1;
	}
	
	if( bank ) {
		s_blk = (bank - 1) * dev_total_blk;
		e_blk = bank * dev_total_blk;
	} else {
		s_blk = 0;
		e_blk = CFG_SD_BANKS * dev_total_blk;
	}
	ret = sd_erase_sect( s_blk, e_blk );
	return ret;
}

/* write */
int 
sd_write_blk( unsigned long addr, unsigned long blk, unsigned int size)
{
	int					ret;
	unsigned long		chk_sblk, chk_eblk;
	struct boot_info	info;
	
	if( addr < PHYS_SDRAM_1 ) {
		return MP200_SD_ERR_PARAM1;
	}
	if( (addr + size) > (PHYS_SDRAM_2 + PHYS_SDRAM_2_SIZE) ) {
		return MP200_SD_ERR_PARAM3;
	}
	
	chk_sblk  = (blk / CFG_SD_SECT_OF_BLOCKS) * CFG_SD_SECT_OF_BLOCKS;
	chk_eblk  = ((size + CFG_SD_SECT_OF_BLOCKS - 1) / CFG_SD_SECT_OF_BLOCKS) * CFG_SD_SECT_OF_BLOCKS;
	chk_eblk += chk_sblk;
	printf("protect check: start = %x, end = %x\n", chk_sblk, chk_eblk);
	ret = sd_check_protect( chk_sblk, chk_eblk, 0 );
	if( ret ) {
		return MP200_SD_ERR_PROTECT;
	}
	
	/* set sd erase information */
	info.addr		= addr;	/* src address */
	info.block		= blk;	/* dst block */
	info.size		= size;	/* write size [byte] */
	ret = sd_write( &info );
	if( ret ) {
		return MP200_SD_ERR_WRITE;
	}

	return ret;
}

int 
sd_write_part( unsigned long addr, unsigned long part, unsigned int size)
{
	int				ret=0;
	unsigned long	blk, p_size;
	MBRecord		*mbr_data = (MBRecord *)ROMBOOT_DATA_START;
	
	blk = mbr_data->partitionTable[part].firstSectorNumbers;
	if( !blk ) {
		return MP200_SD_ERR_PARAM2;
	}
	
	p_size = mbr_data->partitionTable[part].numberOfSectors;
	if( ((size + SDM_BLOCKLEN_VAL - 1) / SDM_BLOCKLEN_VAL) > p_size ) {
		return MP200_SD_ERR_PARAM3;
	}
	ret = sd_write_blk( addr, blk, size );

	return ret;
}

/* protect */
int 
sd_protect_sect( unsigned long s_blk, unsigned long e_blk, int mode )
{
	unsigned char	*wp_tbl = (unsigned char *)CFG_SD_PROTECT_TBL;
	unsigned long	sect, e_sect;
	if( (s_blk & (CFG_SD_SECT_OF_BLOCKS - 1)) || (s_blk >= dev_total_blk) ) {
		return MP200_SD_ERR_PARAM1;
	}
	if( (e_blk & (CFG_SD_SECT_OF_BLOCKS - 1)) || (s_blk >= e_blk) || (e_blk > dev_total_blk) ) {
		return MP200_SD_ERR_PARAM2;
	}
	if( mode & 0xFFFFFFFE ) {
		return MP200_SD_ERR_PARAM3;
	}
	
	sect = s_blk / CFG_SD_SECT_OF_BLOCKS;
	wp_tbl += sect;
	e_sect = (e_blk / CFG_SD_SECT_OF_BLOCKS) - sect;
	if( !e_sect ) {
		e_sect = 1;
	}
	for( sect=0; sect<e_sect; sect++ ) {
		*(wp_tbl++) = mode;
	}
	return sect;
}

int 
sd_protect_bank( unsigned long bank, int mode )
{
	int				ret;
	unsigned long	s_blk, e_blk;
	if( bank > CFG_SD_BANKS ) {
		return MP200_SD_ERR_PARAM1;
	}
	
	if( bank ) {
		s_blk = (bank - 1) * dev_total_blk;
		e_blk = bank * dev_total_blk;
	} else {
		s_blk = 0;
		e_blk = CFG_SD_BANKS * dev_total_blk;
	}

	ret = sd_protect_sect( s_blk, e_blk, mode );
	return ret;
}

static int 
sd_check_protect( unsigned long s_blk, unsigned long e_blk, int print )
{
	unsigned char	*wp_tbl = (unsigned char *)CFG_SD_PROTECT_TBL;
	unsigned long	sect, e_sect, p_blk;
	if( (s_blk & (CFG_SD_SECT_OF_BLOCKS - 1)) || (s_blk >= dev_total_blk) ) {
		return MP200_SD_ERR_PARAM1;
	}
	if( (e_blk & (CFG_SD_SECT_OF_BLOCKS - 1)) || (s_blk >= e_blk) || (e_blk > dev_total_blk) ) {
		return MP200_SD_ERR_PARAM2;
	}
	
	sect = s_blk / CFG_SD_SECT_OF_BLOCKS;
	wp_tbl += sect;
	e_sect = (e_blk / CFG_SD_SECT_OF_BLOCKS) - sect;
	if( print ) {
		p_blk = s_blk;
		for( sect=0; sect<e_sect; sect++ ) {
			printf("%08X ", p_blk);
			p_blk += CFG_SD_SECT_OF_BLOCKS;
			if( *(wp_tbl++) ) {
				printf("(RO) ");
			} else {
				printf("     ");
			}
			if( ((sect+1)%5) == 0 ) {
				printf("\n    ");
			}
		}
		printf("\n");
	} else {
		for( sect=0; sect<e_sect; sect++ ) {
			if( *(wp_tbl++) ) {
				return MP200_SD_ERR_PROTECT;
			}
		}
	}
	return MP200_SD_OK;
}

int 
sd_erase( struct boot_info *boot_info )
{
	int				ret;
	unsigned int	src;
	src  = boot_info->block;
	if( !acc_mode ) {
		src  *= SDM_BLOCKLEN_VAL;
	}
	if( boot_info->erase ) {
		/* erase */
		ret = sd_multi_block_erase( src, boot_info->size );
	} else {
		/* clear */
		ret = sd_multi_block_write( src, (unsigned char *)NULL, boot_info->size );
	}
	return ret;
}

int 
sd_write( struct boot_info *boot_info )
{
	int				ret;
	unsigned int	dest, src;
	
	dest = (unsigned int)boot_info->addr;
	src  = boot_info->block;

	if( !acc_mode ) {
		src  *= SDM_BLOCKLEN_VAL;
	}
	/* write */
	ret = sd_multi_block_write( src, (unsigned char *)dest, boot_info->size );
	return ret;
}

int 
sd_read( struct boot_info *boot_info )
{
	int				ret;
	unsigned int	dest, src;
	dest = (unsigned int)boot_info->addr;
	src  = boot_info->block;
	if( !acc_mode ) {
		src  *= SDM_BLOCKLEN_VAL;
	}
	/* read */
	ret = sd_multi_block_read( src, (unsigned char *)dest, boot_info->size );
	return ret;
}

#endif	/* u-boot */

static int 
sd_send_cmd(int cmd)
{
	int ret=0;
	outl( SDM_INFO_CLEAR, SDM_INFO1 );			/* Clear ALL */
	outl( SDM_INFO_CLEAR, SDM_INFO2 );			/* Clear ALL */
	
	cmd = cmd & 0x0000ffff;
	outl( cmd, SDM_CMD );
	while( !(inl( SDM_INFO1 ) & SDM_INFO1_RES_END) ) {
		if( inl( SDM_INFO2 ) & SDM_INFO2_ERR_ALL ) {
			break;
		}
	}
	ret = inl( SDM_INFO2 ) & SDM_INFO2_ERR_ALL;
	if( ret != 0 ) {
		outl( (inl( SDM_INFO2 ) & ~SDM_INFO2_ERR_ALL), SDM_INFO2 );
	} else {
		outl( (inl(SDM_INFO1) & ~SDM_INFO1_RES_END), SDM_INFO1 );
	}
	return ret;
}

static int
sd_mbr_pram_chk(MBRecord *mbr)
{
	BPBlock *bpb;
	const unsigned char match_fat[5] = {0x46, 0x41, 0x54, 0x31, 0x36}; /* FAT16 */
	/* CHK pattrn */
	if(!((mbr->sig[0] == 0x55) && (mbr->sig[1] == 0xAA))) {
		return -1;
	}
	/* CHK FAT16 (more 32M)*/
	if((mbr->partitionTable[0].fileSystemDescriptor != 0x6)) {
		bpb = (BPBlock *)mbr;
		if(!memcmp(bpb->fileSystemType, match_fat, 5)) {
			return 1;
		}
		return -1;
	}
	return 0;
}

static int
sd_1sector_read(int tSector)
{
	int ret;
	unsigned int address, info2_tmp;
	address = tSector * SDM_BLOCKLEN_VAL;

	outl(address & 0xFFFF, SDM_ARG0);
	outl((address >> 16) & 0xFFFF, SDM_ARG1);
	
	ret = sd_send_cmd(17);
	if(ret != 0){
		return ret;
	}
	while(1) {
		info2_tmp = inl(SDM_INFO2);
		if(info2_tmp & (SDM_INFO2_BRE_BIT | SDM_INFO2_ERR_ALL)){
			break;
		}
	}
	if(info2_tmp & SDM_INFO2_ERR_ALL) {
		return -1;
	}
	return 0;
}

static void
sd_read_data(unsigned char * buf)
{
	int	i;
	unsigned short *tmp_buf;
	tmp_buf = (unsigned short *)buf;
	for(i = 0; i < SDM_HALF_BLOCKLEN_VAL; i++) {
		*tmp_buf++ = (unsigned short)inl(SDM_BUF0);
	}
	return;
}

#if !defined(MP200_MINIBOOT)
static int
sd_multi_block_read( unsigned int address, unsigned char *data, unsigned int size)
{
	int				ret, len, i, j;
	unsigned int	info2_tmp;
	unsigned short *ptr = (unsigned short *)data;
	unsigned int block_num;
	if( (data == NULL) || (size == 0) ) {
		return -1;
	}

	block_num = (size / SDM_BLOCKLEN_VAL);
	if( size & (SDM_BLOCKLEN_VAL-1) ) {
		block_num ++;
	}

	outl( block_num, SDM_SECCNT );
	outl( 0x100, SDM_STOP );

	/* CMD18 : READ_MULTI_BLOCK */
	outl( (address & 0xFFFF),         SDM_ARG0 );
	outl( ((address >> 16) & 0xFFFF), SDM_ARG1 );
	ret = sd_send_cmd( 18 );
	if( ret != 0 ){
		return ret;
	}

	len = inl( SDM_SIZE );
	for ( j = 0; j < block_num; j++ ) {
		while( 1 ) {
			/* wait read enable */
			info2_tmp = inl( SDM_INFO2 );
			if( info2_tmp & (SDM_INFO2_BRE_BIT | SDM_INFO2_ERR_ALL) ) {
				break;
			}
		}
		if( info2_tmp & SDM_INFO2_ERR_ALL ) {
			outl( (info2_tmp & ~SDM_INFO2_ERR_ALL), SDM_INFO2 );
			return -1;
		}
		outl( (info2_tmp & ~SDM_INFO2_BRE_BIT), SDM_INFO2 );
		/* 1block read */
		for( i=0; i<(len/2); i++ ) {
			*ptr++ = inl( SDM_BUF0 );
		}
	}

	/* wait for data read end */
	while( !(inl( SDM_INFO1 ) & SDM_INFO1_RES_RW_END) ) {
		if( inl( SDM_INFO2 ) & SDM_INFO2_ERR_ALL ){
			outl( (inl( SDM_INFO2 ) & ~SDM_INFO2_ERR_ALL), SDM_INFO2 );
			return -1;
		}
	}
	outl( (inl( SDM_INFO1 ) & ~(SDM_INFO1_RES_END | SDM_INFO1_RES_RW_END)), SDM_INFO1 );

	return 0;
}

static int
sd_multi_block_write( unsigned int address, unsigned char *data, unsigned int size )
{
	int				ret, len, i, j;
	unsigned int	info2_tmp;
	unsigned short *ptr = (unsigned short *)data;
	unsigned int	block_num;
	
	if( size == 0 ) {
		return -1;
	}

	block_num = (size / SDM_BLOCKLEN_VAL);
	if( size & (SDM_BLOCKLEN_VAL-1) ) {
		block_num ++;
	}

	outl( block_num, SDM_SECCNT );
	outl( 0x100, SDM_STOP );

	/* CMD25 : WRITE_MULTI_BLOCK */
	outl( (address & 0xFFFF),         SDM_ARG0 );
	outl( ((address >> 16) & 0xFFFF), SDM_ARG1 );
	ret = sd_send_cmd( 25 );
	if( ret != 0 ){
		return ret;
	}

	len = inl( SDM_SIZE );
	for ( j = 0; j < block_num; j++ ) {
		while( 1 ) {
			/* wait write enable */
			info2_tmp = inl( SDM_INFO2 );
			if( info2_tmp & (SDM_INFO2_BWE_BIT | SDM_INFO2_ERR_ALL) ) {
				break;
			}
		}
		if( info2_tmp & SDM_INFO2_ERR_ALL ) {
			outl( (info2_tmp & ~SDM_INFO2_ERR_ALL), SDM_INFO2 );
			return -1;
		}

		outl( (info2_tmp & ~SDM_INFO2_BWE_BIT), SDM_INFO2 );
		if( data != NULL ) {
			/* 1block write */
			for( i=0; i<(len/2); i++ ) {
				outl( *ptr++, SDM_BUF0 );
			}
		} else {
			/* 1block clear */
			for( i=0; i<(len/2); i++ ) {
				outl( 0, SDM_BUF0 );
			}
		}
	}

	/* wait for data read end */
	while( !(inl( SDM_INFO1 ) & SDM_INFO1_RES_RW_END) ) {
		if( inl( SDM_INFO2 ) & SDM_INFO2_ERR_ALL ){
			outl( (inl( SDM_INFO2 ) & ~SDM_INFO2_ERR_ALL), SDM_INFO2 );
			return -1;
		}
	}
	outl( (inl( SDM_INFO1 ) & ~(SDM_INFO1_RES_END | SDM_INFO1_RES_RW_END)), SDM_INFO1 );
	return 0;
}

static int
sd_multi_block_erase( unsigned int address, unsigned int size )
{
	int				ret;
	unsigned int	end_block;

	if( size == 0 ) {
		return -1;
	}
	if( acc_mode ) {
		end_block = address + (size / SDM_BLOCKLEN_VAL) - 1;
		if( size & (SDM_BLOCKLEN_VAL-1) ) {
			end_block ++;
		}
	} else {
		end_block = address + size -1 ;
	}

	/* CMD32 : ERASE_WR_BLK_START */
	outl( (address & 0xFFFF),         SDM_ARG0 );
	outl( ((address >> 16) & 0xFFFF), SDM_ARG1 );
	ret = sd_send_cmd( 32 );
	if( ret != 0 ){
		return ret;
	}

	/* CMD33 : ERASE_WR_BLK_END */
	outl( (end_block & 0xFFFF),         SDM_ARG0 );
	outl( ((end_block >> 16) & 0xFFFF), SDM_ARG1 );
	ret = sd_send_cmd( 33 );
	if( ret != 0 ){
		return ret;
	}

	/* CMD38 : ERASE */
	outl( 0, SDM_ARG0 );
	outl( 0, SDM_ARG1 );
	ret = sd_send_cmd( 38 );
	if( ret != 0 ){
		return ret;
	}
	/* wait DATA0 == 1 */
	while( !(inl( SDM_INFO2 ) & SDM_INFO2_DATA0_BIT) ) {
	}
	return 0;
}

#else	/* defined(MP200_MINIBOOT) */
static int
sd_bpb_pram_chk(BPBlock *bpb)
{
	/* CHK pattrn */
	if(!((bpb->sig[0] == 0x55) && (bpb->sig[1] == 0xAA))) {
		return -1;
	}
	if((bpb->bytesPerSector != 0x200)) {
		return -1;
	}
	return 0;
}

const unsigned char match_name[3][8] = {{0x75, 0x62, 0x6F, 0x6F, 0x74, 0x2D, 0x73, 0x64},  /* uboot-sd */
                                        {0x75, 0x49, 0x6D, 0x61, 0x67, 0x65, 0x20, 0x20},  /* uImage */
                                        {0x63, 0x72, 0x61, 0x6D, 0x66, 0x73, 0x20, 0x20}}; /* cramfs */
const unsigned char MATCH_NAME[3][8] = {{0x55, 0x42, 0x4F, 0x4F, 0x54, 0x2D, 0x53, 0x44},  /* UBOOT-SD */
                                        {0x55, 0x49, 0x4D, 0x41, 0x47, 0x45, 0x20, 0x20},  /* UIMAGE */
                                        {0x43, 0x52, 0x41, 0x4D, 0x46, 0x53, 0x20, 0x20}}; /* CRAMFS */

static unsigned short
sd_get_first_cluster(unsigned char *read_data, sd_info_t *sd_info, int type)
{
	int	i,j,ret;
	RDEntry *rde;

	const unsigned char match_exet[3] = {0x62, 0x69, 0x6E}; /* bin */
	const unsigned char MACTH_EXET[3] = {0x42, 0x49, 0x4E}; /* BIN */
	for(i = 0; i < sd_info->sec_rde_size; i++) {
		ret = sd_1sector_read(sd_info->rde_sec_pos + i);
		if(ret != 0) {
			dbg_print("read ERR\n");
			return ret;
		}
		sd_read_data(read_data);
		for(j = 0; j < 16; j++) {
			rde = (RDEntry *)(read_data + (32 * j));
			if(!memcmp(rde->name, match_name[type], 8) || !memcmp(rde->name, MATCH_NAME[type], 8)) {
				if(type != LOAD_TYPE_UBOOT) {
					sd_info->data_num_clus = rde->cluster;
					return 0;
				} else if(!memcmp(rde->extension, match_exet, 3) || !memcmp(rde->extension, MACTH_EXET, 3)) {
					sd_info->data_num_clus = rde->cluster;
					return 0;
				}
			}
		}
	}
	return -1;
}

static int
sd_multi_read(sd_info_t *sd_info, int tSector, unsigned char *boot_addr)
{
	unsigned int address, info2_tmp;
	int ret, i, j;
	unsigned short *tmp_buf;
	tmp_buf = (unsigned short *)boot_addr;

	outl(inl(SDM_STOP) | SDM_STOP_SEC_BIT, SDM_STOP);
	outl(sd_info->sec_per_clus, SDM_SECCNT);

	address = tSector * SDM_BLOCKLEN_VAL;

	outl(address & 0xFFFF, SDM_ARG0);
	outl((address >> 16) & 0xFFFF, SDM_ARG1);
	
	ret = sd_send_cmd(18);
	if(ret != 0) {
		return ret;
	}
	for(i = 0 ; i < sd_info->sec_per_clus; i++) {
		while(1) {
			info2_tmp = inl(SDM_INFO2);
			if(info2_tmp & (SDM_INFO2_BRE_BIT | SDM_INFO2_ERR_ALL)) {
				break;
			}
		}
		if(info2_tmp & SDM_INFO2_ERR_ALL) {
			return -1;
		}
		outl(inl(SDM_INFO2) & ~SDM_INFO2_BRE_BIT, SDM_INFO2);
		for(j = 0; j < SDM_HALF_BLOCKLEN_VAL; j++) {
			*tmp_buf++ = (unsigned short)inl(SDM_BUF0);
		}
	}
	return 0;
}

static int
sd_cpy_data(unsigned char *read_data, sd_info_t *sd_info, unsigned int start_addr, unsigned int max_size)
{
	int	ret, last_flag = 0, write_size = 0, clus_size;
	int	read_clus_sec, fat_sec_num, get_clus;
	unsigned short *read_fat_buf = (unsigned short *)read_data;
	unsigned char *boot_addr = (volatile unsigned char *) start_addr;
	
	clus_size = SDM_BLOCKLEN_VAL * sd_info->sec_per_clus;
	while((write_size + clus_size) <= max_size) {
		read_clus_sec = sd_info->user_sec_pos + 
							((sd_info->data_num_clus - 2) * sd_info->sec_per_clus);
		ret = sd_multi_read(sd_info, read_clus_sec, boot_addr + write_size);
		if(ret != 0) {
			return ret;
		}
		fat_sec_num = sd_info->data_num_clus >> 8;
		if(sd_info->now_fat_sec != fat_sec_num) {
			ret = sd_1sector_read(sd_info->fat_sec_pos + fat_sec_num);
			if(ret != 0) {
				return ret;
			}
			sd_read_data(read_data);
			sd_info->now_fat_sec = fat_sec_num;
		}
		get_clus = read_fat_buf[sd_info->data_num_clus & 0xFF];
		if(get_clus < 0x0002 || get_clus > 0xFFF6) {
			if(get_clus >= 0xFFF8 && get_clus <= 0xFFFF) {
				last_flag = 1;
			}
			break;
		}
		sd_info->data_num_clus = get_clus;
		write_size += clus_size;
	}
	if(last_flag != 1) {
		if(SDM_BOOT_MAX_LOADSIZE == max_size ){
			dbg_print(" uboot-sd.bin is over 512KB.\n");
		}else{
			dbg_print(" uImage is over 3.5MB.\n");
		}
		return -1;
	}
	return 0;
}

void
mini_boot(void)
{
	int				ret, ret2;
	int				*acc_mode = (int *)SRAM_ACC_MODE;
	MBRecord		*mbr;
	BPBlock			*bpb;
	sd_info_t		sd_info;
	unsigned char	*read_data = (volatile unsigned char *)ROMBOOT_DATA_START;
	unsigned int	RCA_VAL;
	
	uart_init();

	/* byte mode only */
	*acc_mode = 0;

	/* CMD7 : DESELECT_CARD */
	outl(0, SDM_ARG0);
	outl(0, SDM_ARG1);
	ret = sd_send_cmd(7);		 /* CMD7 */
	/* wait DATA0 == 1 */
	while( !(inl( SDM_INFO2 ) & SDM_INFO2_DATA0_BIT) ) {
	}

	/* CMD3 SET */
	outl(0, SDM_ARG0);
	outl(0, SDM_ARG1);

	ret = sd_send_cmd(3);		 /* CMD3 */
	if(ret != 0) {
		dbg_print("CMD3 error\n");
		goto ERR_END;
	}
	RCA_VAL = inl(SDM_RSP1);

	/* CMD7 : SELECT */
	outl(0, SDM_ARG0);
	outl(RCA_VAL, SDM_ARG1);
	ret = sd_send_cmd(7);		 /* CMD7 */
	if( ret != 0 ){
		dbg_print("CMD7 error\n");
		goto ERR_END;
	}
	
	/* CMD55 SET */
	outl(0x0000, SDM_ARG0);
	outl(RCA_VAL, SDM_ARG1);
	ret = sd_send_cmd(55);		/* CMD55 */
	if(ret != 0) {
		dbg_print("CMD55 error\n");
		goto ERR_END;
	}

	/* CMD6 SET Width 4bit */
	outl(0x0002, SDM_ARG0);
	outl(0x0000, SDM_ARG1);
	ret = sd_send_cmd(0x46); 	/* CMD6 */
	if(ret != 0) {
		dbg_print("CMD6 error\n");
		goto ERR_END;
	}
	/* set Width 4bit */
	outl((SDM_OPTION_VAL & ~SDM_OPT_WIDTH_BIT), SDM_OPTION);

	mbr = (MBRecord *)read_data;
		ret = sd_1sector_read(0);
	if(ret != 0) {
		dbg_print("read error\n");
		goto ERR_END;
	}
	sd_read_data(read_data);

	ret = sd_mbr_pram_chk(mbr);
	if(ret == -1) {
		dbg_print("check MasterBootRecord error\n");
		goto ERR_END;
	} else if(ret == 0) {
		sd_info.bpb_sec_pos = mbr->partitionTable[0].firstSectorNumbers;
		ret2 = sd_1sector_read(sd_info.bpb_sec_pos);
		if(ret2 != 0) {
			dbg_print("read error\n");
			goto ERR_END;
		}
		sd_read_data(read_data);
	} else if(ret == 1) {
		sd_info.bpb_sec_pos = 0;
	}
	bpb = (BPBlock *)read_data;
	ret = sd_bpb_pram_chk(bpb);
	if(ret != 0) {
		dbg_print("check BIOS Parameter Block error\n");
		goto ERR_END;
	}
	sd_info.fat_sec_pos = sd_info.bpb_sec_pos + bpb->reservedSectors;
	sd_info.rde_sec_pos = sd_info.fat_sec_pos + (bpb->numberOfFATs * bpb->sectorsPerFAT);
	sd_info.user_sec_pos = sd_info.rde_sec_pos + ((bpb->rootEntries * 32) >> 9);

	sd_info.byte_per_sec = bpb->bytesPerSector;
	sd_info.sec_per_clus = bpb->sectorsPerCluster;
	sd_info.root_entry = bpb->rootEntries;
	sd_info.sec_per_fat = bpb->sectorsPerFAT;
	sd_info.sec_rde_size = (sd_info.root_entry * 32) >> 9;
#ifdef CONFIG_MP200_SDBOOT_LINE
	dbg_print("\n\nLoad LineSystem U-BOOT ... ");
#endif
	/* uboot-sd.bin */
	ret = sd_get_first_cluster(read_data, &sd_info, LOAD_TYPE_UBOOT);
	if(ret != 0) {
		dbg_print("file uboot-sd.bin not exist error\n");
		goto ERR_END;
	}
	sd_info.now_fat_sec = -1;

	ret = sd_cpy_data(read_data, &sd_info, (unsigned int)U_BOOT_START, SDM_BOOT_MAX_LOADSIZE);
	if(ret != 0) {
		dbg_print("copy uboot-sd.bin error\n");
		goto ERR_END;
	}

#ifdef CONFIG_MP200_SDBOOT_LINE
	dbg_print("Done\n");
	dbg_print("Load LineSystem Kernel ... ");
#endif
	/* uImage */
	ret = sd_get_first_cluster(read_data, &sd_info, LOAD_TYPE_KERNEL);
	if(ret != 0) {
		dbg_print("file uImage not exist error\n");
		goto ERR_END;
	}
	sd_info.now_fat_sec = -1;

	ret = sd_cpy_data(read_data, &sd_info, (unsigned int)KERNEL_START, KERNEL_SIZE);
	if(ret != 0) {
		dbg_print("copy uImage error\n");
		goto ERR_END;
	}
	
#ifdef CONFIG_MP200_SDBOOT_LINE
	dbg_print("Done\n");
	dbg_print("Load LineSystem Filesystem ... ");

	/* cramfs */
	ret = sd_get_first_cluster(read_data, &sd_info, LOAD_TYPE_FS);
	if(ret != 0) {
		dbg_print("\nfile cramfs not exist error\n");
		goto READ_SKIP;
	}
	sd_info.now_fat_sec = -1;

	ret = sd_cpy_data(read_data, &sd_info, (unsigned int)FS_START, FS_SIZE);
	if(ret != 0) {
		dbg_print("\ncopy cramfs error\n");
		goto READ_SKIP;
	}

	dbg_print("Done\n");
 READ_SKIP:
#endif
	
	/* jump mini-boot */
	((void (*)(void))U_BOOT_START) ();

 ERR_END:
	outl(0x0020, SDM_CLK_CTRL); /* Clk Disable */
	return;
}

#endif	/* defined(MP200_MINIBOOT) */

#endif	/* defined(CONFIG_MP200_SDBOOT) */

