/*******************************************************************************
* Copyright (C) NEC Electronics Corporation 2008
* NEC ELECTRONICS CONFIDENTIAL AND PROPRIETARY
* All rights reserved by NEC Electronics Corporation.
* Use of copyright notice does not evidence publication
*
*  File Name : sd/sd.h
*  Contents  : sd header
*  History   :
*   Date        Rev.      Who         Details     Keyword
*  ==========  ========  ==========  ==========  ==========
*  2009/ 2/19  Rev1.01   k-kuri
*
*******************************************************************************/

/* register address define */
/* base address */
#define	EM1_SD0_BASE			0x50050000

#define SDM_CMD					(EM1_SD0_BASE + 0x0000)
#define SDM_PORTSEL				(EM1_SD0_BASE + 0x0004)
#define SDM_ARG0				(EM1_SD0_BASE + 0x0008)
#define SDM_ARG1				(EM1_SD0_BASE + 0x000c)
#define SDM_STOP				(EM1_SD0_BASE + 0x0010)
#define SDM_SECCNT				(EM1_SD0_BASE + 0x0014)
#define SDM_RSP0				(EM1_SD0_BASE + 0x0018)
#define SDM_RSP1				(EM1_SD0_BASE + 0x001c)
#define SDM_RSP2				(EM1_SD0_BASE + 0x0020)
#define SDM_RSP3				(EM1_SD0_BASE + 0x0024)
#define SDM_RSP4				(EM1_SD0_BASE + 0x0028)
#define SDM_RSP5				(EM1_SD0_BASE + 0x002c)
#define SDM_RSP6				(EM1_SD0_BASE + 0x0030)
#define SDM_RSP7				(EM1_SD0_BASE + 0x0034)

#define SDM_INFO1				(EM1_SD0_BASE + 0x0038)
#define SDM_INFO2				(EM1_SD0_BASE + 0x003c)
#define SDM_INFO1_MASK			(EM1_SD0_BASE + 0x0040)
#define SDM_INFO2_MASK			(EM1_SD0_BASE + 0x0044)
#define SDM_CLK_CTRL			(EM1_SD0_BASE + 0x0048)
#define SDM_SIZE				(EM1_SD0_BASE + 0x004C)

#define SDM_OPTION				(EM1_SD0_BASE + 0x0050)

#define SDM_BUF0				(EM1_SD0_BASE + 0x0060)

#define SDM_SOFT_RST			(EM1_SD0_BASE + 0x01c0)

/* register set value define */
#define SDM_SOFT_RST_MDL		0x00000000	/* module reset */
#define SDM_SOFT_RST_SFT		0x00000007	/* release reset */
#define SDM_OPTION_VAL			0x000080E8	/* response time out counter */
#define SDM_INFO_CLEAR			0x00000000	/* interrupt status clear */
#define SDM_INFO_MASK_ALL		0x0000ffff	/* interrupt mask */
#define SDM_INFO_MASK_FREE		0x00000000	/* interrupt enable */
#define SDM_CLK_CTRL_DIV256		0x00000140	/* enable SD clock(LBUS(=57.344) * 1/256) no control */
#define SDM_CLK_CTRL_DIV8		0x00000302	/* enable SD clock(LBUS(=57.344) * 1/  8) auto control */
#define SDM_CLK_CTRL_DIV4		0x00000301	/* enable SD clock(LBUS(=57.344) * 1/  4) auto control */
#define SDM_CLK_CTRL_DIV2		0x00000700	/* enable SD clock(LBUS(=83.3) * 1/  2) auto control */
#define SDM_INFO1_MASK_VAL		0x0000fffa	/* Response end,R/W access all end Enable */
#define SDM_INFO2_MASK_VAL		0x0000807F	/*  */

#define SDM_INFO1_RES_END		0x0001		/* RES END BIT */
#define SDM_INFO1_RES_RW_END	0x0004      /* RES Read/Write END BIT */

#define SDM_INFO2_ERR_ALL		0x807F		/* INFO2 ERR */
#define SDM_INFO2_BWE_BIT		0x0200      /* INFO2 BWE BIT */
#define SDM_INFO2_BRE_BIT		0x0100		/* INFO2 BRE BIT */
#define SDM_INFO2_DATA0_BIT		0x0080		/* INFO2 DATA0 BIT */
#define SDM_INFO2_RES_TOUT		0x0040		/* INFO2 ERR6 BIT */

#define SDM_BLOCKLEN_VAL		512
#define SDM_HALF_BLOCKLEN_VAL	256

#define SDM_OPT_WIDTH_BIT		0x8000

#define SDM_STOP_SEC_BIT		0x0100

#define MBR_ADDR				0			/* 0x0000-0x01FF */
#define MBR_PTABLE_NUM			4			/* Number of Partition Table in MBR */
#define MBR_ACTIVE_PT			0x80		/* Active Partition Table */

#define LOAD_TYPE_UBOOT			0
#define LOAD_TYPE_KERNEL		1
#define LOAD_TYPE_FS			2

#define U_BOOT_START			0x31080000
#define SDM_BOOT_MAX_LOADSIZE	0x00040000	/* 256k */

#define KERNEL_START			CFG_LOAD_ADDR	/* SDRAM */
#define KERNEL_SIZE				0x00380000	/* 3.5M */

#define FS_START				0x34000000	/* SDRAM */
#define FS_SIZE					0x02000000	/* 32M */


typedef struct MasterBootRecord {
	unsigned char nondata[446];
	struct {
		unsigned char	bootDescriptor;
		unsigned char	firstPartitionSector[3];
		unsigned char	fileSystemDescriptor;
		unsigned char	lastPartitionSector[3];
		unsigned int	firstSectorNumbers;
		unsigned int	numberOfSectors;
	} partitionTable[4];
	unsigned char    sig[2]; 
} __attribute__ ((packed)) MBRecord;

typedef struct BIOS_Parameter_Block {
	unsigned char	jmpOpeCode[3];
	unsigned char	OEMName[8];
	unsigned short	bytesPerSector;
	unsigned char	sectorsPerCluster;
	unsigned short	reservedSectors;
	unsigned char	numberOfFATs;
	unsigned short	rootEntries;
	unsigned short	totalSectors;
	unsigned char	mediaDescriptor;
	unsigned short	sectorsPerFAT;
	unsigned short	sectorsPerTrack;
	unsigned short	heads;
	unsigned int	hiddenSectors;
	unsigned int	bigTotalSectors;
	/* info */
	unsigned char	driveNumber;
	unsigned char	unused;
	unsigned char	extBootSignature;
	unsigned int	serialNumber;
	unsigned char	volumeLabel[11];
	unsigned char	fileSystemType[8];
	unsigned char	loadProgramCode[448];
	unsigned char	sig[2];
} __attribute__ ((packed)) BPBlock;

typedef struct Root_Directory_Entry{
	unsigned char	name[8];
	unsigned char	extension[3];
	unsigned char	attribute;
	unsigned char	reserved;
	unsigned char	createTimeMs;
	unsigned short	createTime;
	unsigned short	createDate;
	unsigned short	accessDate;
	unsigned short	clusterHighWord;
	unsigned short	updateTime;
	unsigned short	updateDate;
	unsigned short	cluster;
	unsigned int	fileSize;
} __attribute__ ((packed)) RDEntry;

typedef struct SD_INFO{
	int	bpb_sec_pos;
	int	fat_sec_pos;
	int	rde_sec_pos;
	int	user_sec_pos;
	int	byte_per_sec;
	int	sec_per_clus;
	int	root_entry;
	int	sec_per_fat;
	int	sec_rde_size;
	int	data_num_clus;
	int	now_fat_sec;
} sd_info_t;

