/*
 * (C) Copyright 2004
 * Jian Zhang, Texas Instruments, jzhang@ti.com.

 * (C) Copyright 2000-2006
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * (C) Copyright 2001 Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Andreas Heppel <aheppel@sysgo.de>

 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/* #define DEBUG */

#include <common.h>

#if defined(CFG_ENV_IS_IN_SD)

#include <command.h>
#include <environment.h>
#include <linux/stddef.h>
#include <malloc.h>

#if ((CONFIG_COMMANDS&(CFG_CMD_ENV|CFG_CMD_SD)) == (CFG_CMD_ENV|CFG_CMD_SD))
#define CMD_SAVEENV
#elif defined(CFG_ENV_OFFSET_REDUND)
#error Cannot use CFG_ENV_OFFSET_REDUND without CFG_CMD_ENV & CFG_CMD_NAND
#endif

#if defined(CFG_ENV_SIZE_REDUND) && (CFG_ENV_SIZE_REDUND != CFG_ENV_SIZE)
#error CFG_ENV_SIZE_REDUND should be the same as CFG_ENV_SIZE
#endif

#ifdef CONFIG_INFERNO
#error CONFIG_INFERNO not supported yet
#endif

/* references to names in env_common.c */
extern uchar default_environment[];
extern int default_environment_size;

char * env_name_spec = "SD Memory Card";

env_t *env_ptr = (env_t *)CFG_ENV_TMP_ADDR;
static unsigned long env_offset = 0;

struct sd_rw_info {
	unsigned long addr;		/* copy destination address */
	unsigned long block;	/* copy source block number */
	unsigned long size;		/* copy size [byte] */
	unsigned long erase;	/* erase flag */
};

/* extern function */
extern unsigned long sd_flash_init( void );
extern int sd_read( struct sd_rw_info * );
#ifdef CMD_SAVEENV
extern int sd_write( struct sd_rw_info * );
#endif
extern int sd_erase( struct sd_rw_info * );

/* extern variable */
extern uchar default_environment[];
extern int default_environment_size;


uchar env_get_char_spec( int index )
{
	DECLARE_GLOBAL_DATA_PTR;

	return ( *((uchar *)(gd->env_addr + index)) );
}

int env_init( void )
{
	DECLARE_GLOBAL_DATA_PTR;
	int ret = -1;
	struct sd_rw_info read_info;
	
	/* initialize protect table */
	env_offset = sd_flash_init();
	
	if( env_offset ) {
		/* set nand read information */
		read_info.addr		= CFG_ENV_TMP_ADDR;	/* copy destination address */
		read_info.block		= env_offset;		/* copy source block number */
		read_info.size		= CFG_ENV_SIZE;		/* copy size [block] */
		ret = sd_read( &read_info );
	}
	if( !ret && (crc32(0, env_ptr->data, ENV_SIZE) == env_ptr->crc) ) {
		gd->env_addr  = (ulong)&(env_ptr->data);
		gd->env_valid = 1;
	} else {
		gd->env_addr  = (ulong)&default_environment[0];
		gd->env_valid = 0;
	}
	
	return (0);
}

#ifdef CMD_SAVEENV
int saveenv( void )
{
	int ret;
	struct sd_rw_info sd_info;
	
	if( !env_offset ) {
		puts ("error: no area to save enviroment information in SD Memory Card\n");
		return 1;
	}
	/* copy current environment to temporary buffer */
	memcpy( (uchar *)CFG_ENV_TMP_ADDR, env_ptr, CFG_ENV_SIZE );

#if 0
	puts ("Erasing to SD Memory Card... ");
	/* set nand erase information */
	sd_info.block		= env_offset;		/* erase block number */
	sd_info.size		= CFG_ENV_SIZE;		/* erase size [byte] */
	sd_info.erase		= 0;				/* erase flag */
	ret = sd_erase( &sd_info );
	if( !ret ) {
		puts ("done\n");
	} else {
		puts ("error\n");
		return 1;
	}
#endif

	puts ("Writing to SD Memory Card... ");
	/* set nand write information */
	sd_info.addr		= CFG_ENV_TMP_ADDR;	/* copy destination address */
	sd_info.block		= env_offset;		/* copy source block number */
	sd_info.size		= CFG_ENV_SIZE;		/* copy size [byte] */
	ret = sd_write( &sd_info );
	if( !ret ) {
		puts ("done\n");
		return 0;
	} else {
		puts ("error\n");
		return 1;
	}
}

int initenv( void )
{
	int ret;
	struct sd_rw_info sd_info;
	
	if( !env_offset ) {
		puts ("error: no area to save enviroment information in SD Memory Card\n");
		return 1;
	}

	puts ("Erasing to SD Memory Card... ");
	/* set nand erase information */
	sd_info.block		= env_offset;		/* erase block number */
	sd_info.size		= CFG_ENV_SIZE;		/* erase size [byte] */
	sd_info.erase		= 0;				/* erase flag */
	ret = sd_erase( &sd_info );
	if( !ret ) {
		puts ("done\n");
	} else {
		puts ("error\n");
		return 1;
	}
}

#endif /* CMD_SAVEENV */

void env_relocate_spec( void )
{
	memcpy (env_ptr, (void*)CFG_ENV_TMP_ADDR, CFG_ENV_SIZE);
}

#endif /* CFG_ENV_IS_IN_SD */
